#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "php.h"
#include "main/php_network.h"

#include <gnutls/gnutls.h>
#include "php_gnutls.h"

typedef struct {
        php_socket_t bsd_socket;
        int             type;
        int             error;
} php_socket;


static int le_gnutls_session;
#define le_gnutls_session_name  "GnuTLS Resource"


void gnutls_session_destruction_handler(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{

	gnutls_session_t *my_rsrc = (gnutls_session_t *) rsrc->ptr;
	gnutls_deinit(*my_rsrc);

	efree(my_rsrc);
}


static int le_gnutls_credentials;
#define le_gnutls_credentials_name  "GnuTLS Credentials"

void gnutls_credentials_destruction_handler(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{

	gnutls_credentials_t *my_rsrc = (gnutls_credentials_t *) rsrc->ptr;


	if(my_rsrc->type == CRED_ANON_SERVER)
		gnutls_anon_free_server_credentials((gnutls_anon_server_credentials_t )my_rsrc->credentials);
	if(my_rsrc->type == CRED_ANON_CLIENT)
		gnutls_anon_free_client_credentials((gnutls_anon_client_credentials_t )my_rsrc->credentials);	

	efree(my_rsrc->credentials);
	efree(my_rsrc);
}





/* compiled function list so Zend knows what's in this module */
zend_function_entry gnutls_functions[] =
{
	ZEND_FE(gnutls_session_new, NULL)
	ZEND_FE(gnutls_session_set_socket, NULL)
	ZEND_FE(gnutls_credentials_new, NULL)
	ZEND_FE(gnutls_session_set_credentials, NULL)
	ZEND_FE(gnutls_set_default_priority, NULL)
	ZEND_FE(gnutls_kx_set_priority, NULL)
	ZEND_FE(gnutls_handshake, NULL)
	ZEND_FE(gnutls_record_recv, NULL)
	ZEND_FE(gnutls_record_send, NULL)
	ZEND_FE(gnutls_bye, NULL)
	ZEND_FE(gnutls_session_get_data, NULL)
	ZEND_FE(gnutls_session_set_data, NULL)
	ZEND_FE(gnutls_session_is_resumed, NULL)
	{NULL, NULL, NULL}
};

/* compiled module information */
zend_module_entry gnutls_module_entry =
{
    STANDARD_MODULE_HEADER,
    "gnutls Module",
    gnutls_functions,
    ZEND_MINIT(gnutls_module_init), 
    ZEND_MSHUTDOWN(gnutls_module_shutdown), 
    NULL, 
    NULL, 
    NULL,
    NO_VERSION_YET,
    STANDARD_MODULE_PROPERTIES
};

#ifdef DEBUG

void gnutls_php_log(int level, const char *msg)
{

	zend_printf("[%d] [%s]\n", level, msg);

}
#endif

/* Init/Deinit */

ZEND_MINIT_FUNCTION(gnutls_module_init) 
{


	gnutls_global_init ();

        #ifdef DEBUG
	       gnutls_global_set_log_level(10);
                gnutls_global_set_log_function(&gnutls_php_log);
        #endif

	le_gnutls_session = zend_register_list_destructors_ex(gnutls_session_destruction_handler, NULL, le_gnutls_session_name, module_number);
	le_gnutls_credentials = zend_register_list_destructors_ex(gnutls_credentials_destruction_handler, NULL, le_gnutls_credentials_name, module_number);

	REGISTER_LONG_CONSTANT("GNUTLS_KX_ANON_DH", GNUTLS_KX_ANON_DH, CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("GNUTLS_CLIENT", GNUTLS_CLIENT, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("GNUTLS_SERVER", GNUTLS_SERVER, CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("CRED_ANON_CLIENT", CRED_ANON_CLIENT, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("CRED_ANON_SERVER", CRED_ANON_SERVER, CONST_CS | CONST_PERSISTENT);


	REGISTER_LONG_CONSTANT("GNUTLS_SHUT_RDWR", GNUTLS_SHUT_RDWR, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("GNUTLS_SHUT_WR", GNUTLS_SHUT_WR, CONST_CS | CONST_PERSISTENT);



	return SUCCESS;
}

ZEND_MSHUTDOWN_FUNCTION(gnutls_module_shutdown)
{
	gnutls_global_deinit ();

	//zend_list_delete(le_gnutls_session);
	//zend_list_delete(le_gnutls_credentials);

	return SUCCESS;
}


/* implement standard "stub" routine to introduce ourselves to Zend */

ZEND_GET_MODULE(gnutls)


/* implement function that is meant to be made available to PHP */
ZEND_FUNCTION(gnutls_session_new)
{
	void *session;
	unsigned long flag;
	int ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &flag) == FAILURE) 
	{
	    return;
	}

	session = emalloc(sizeof(gnutls_session_t));


	ret = gnutls_init ((gnutls_session_t *)session, flag);


	ZEND_REGISTER_RESOURCE(return_value, session, le_gnutls_session);

}

ZEND_FUNCTION(gnutls_credentials_new)
{
	gnutls_credentials_t *credentials;
	unsigned long flag;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &flag) == FAILURE) 
	{
	    return;
	}

	credentials = emalloc(sizeof(gnutls_credentials_t));

	switch(flag)
	{
		case CRED_ANON_SERVER:
				credentials->type = CRED_ANON_SERVER;
				credentials->credentials = emalloc(sizeof(gnutls_anon_server_credentials_t));
				gnutls_anon_allocate_server_credentials((gnutls_anon_server_credentials_t *)credentials->credentials);
				break;
		case CRED_ANON_CLIENT:
				credentials->type = CRED_ANON_CLIENT;
				credentials->credentials = emalloc(sizeof(gnutls_anon_client_credentials_t));
				gnutls_anon_allocate_client_credentials((gnutls_anon_client_credentials_t *)credentials->credentials);
				break;
		default:
				credentials->type = 0;
	}


	ZEND_REGISTER_RESOURCE(return_value, credentials, le_gnutls_credentials);

}

ZEND_FUNCTION(gnutls_session_set_credentials)
{
	zval *session_arg, *credentials_arg;
	gnutls_session_t *session;
	gnutls_credentials_t *credentials;
	int ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr", &session_arg, &credentials_arg) == FAILURE) 
	{
	    return;
	}	

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);
	ZEND_FETCH_RESOURCE(credentials, gnutls_credentials_t *, &credentials_arg, -1, le_gnutls_credentials_name, le_gnutls_credentials);

	if(credentials->type == CRED_ANON_CLIENT)
	{
		ret = gnutls_credentials_set (*session, GNUTLS_CRD_ANON, *((gnutls_anon_client_credentials_t *)credentials->credentials));
	}
	else
	if(credentials->type == CRED_ANON_SERVER)
	{
		ret = gnutls_credentials_set (*session, GNUTLS_CRD_ANON, *((gnutls_anon_server_credentials_t *)credentials->credentials));
	}

	RETURN_LONG(ret);
}





ZEND_FUNCTION(gnutls_session_set_socket)
{
	zval *session_arg, *socket_arg;
	php_socket	*php_sock;
	gnutls_session_t *session;

	int le_socket = 0;
	char *le_socket_name = "Socket";


	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr", &session_arg, &socket_arg) == FAILURE) 
	{
	    return;
	}	
	
	ZEND_GET_RESOURCE_TYPE_ID(le_socket, le_socket_name);


	ZEND_FETCH_RESOURCE(php_sock, php_socket *, &socket_arg, -1, le_socket_name, le_socket);
	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);



	gnutls_transport_set_ptr (*session, (gnutls_transport_ptr_t) php_sock->bsd_socket);

}

ZEND_FUNCTION(gnutls_set_default_priority)
{
	zval *session_arg;
	gnutls_session_t *session;
	int ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_arg) == FAILURE) 
	{
	    return;
	}	

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);


	ret = gnutls_set_default_priority(*session);	


	RETURN_LONG(ret);

}


ZEND_FUNCTION(gnutls_kx_set_priority)
{
	zval *session_arg;
	gnutls_session_t *session;
	long priority_arg;
	int *prioritys = emalloc(sizeof(int) * 2);
	int ret;


	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &session_arg, &priority_arg) == FAILURE) 
	{
	    return;
	}	

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);

	prioritys[0] = priority_arg;
	prioritys[1] = 0;

	ret = gnutls_kx_set_priority(*session, prioritys);	

	efree(prioritys);

	RETURN_LONG(ret);

}



ZEND_FUNCTION(gnutls_handshake)
{
	zval *session_arg;
	gnutls_session_t *session;
	int ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_arg) == FAILURE) 
	{
	    return;
	}	

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);


	ret = gnutls_handshake(*session);	


	RETURN_LONG(ret);
}


ZEND_FUNCTION(gnutls_record_recv)
{
	zval *session_arg;
	char *buf;
	gnutls_session_t *session;

	int retval;
	long length;
	


	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &session_arg, &length) == FAILURE) 
	{
	    return;
	}

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);

	if (length < 1) 
	{
		RETURN_FALSE;
	}


	buf = emalloc(length + 1);

	
	retval = gnutls_record_recv (*session, buf, length);


	if(retval > 0)
	{
		buf = erealloc(buf, retval + 1);
			
		buf[retval] = 0;
		
		RETURN_STRINGL(buf, retval, 0);
	}

	
	efree(buf);
	RETURN_EMPTY_STRING();

}

ZEND_FUNCTION(gnutls_record_send)
{
	zval *session_arg;
	gnutls_session_t *session;
	char *string;
	int string_len, retval;
	long length;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|l", &session_arg, &string, &string_len, &length) == FAILURE) 
	{
	    return;
	}	

	if (ZEND_NUM_ARGS() < 3) 
	{
		length = string_len;
	}

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);


	retval = gnutls_record_send (*session, string, length);

	RETURN_LONG(retval);
}

ZEND_FUNCTION(gnutls_bye)
{
	zval *session_arg;
	gnutls_session_t *session;
	long flag, retval;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &session_arg, &flag) == FAILURE) 
	{
	    return;
	}

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);

	retval = gnutls_bye(*session, flag);

	RETURN_LONG(retval);
}



ZEND_FUNCTION(gnutls_session_is_resumed)
{
	zval *session_arg;
	gnutls_session_t *session;
	long retval;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_arg) == FAILURE) 
	{
	    return;
	}

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);

	retval = gnutls_session_is_resumed(*session);

	RETURN_LONG(retval);

}


ZEND_FUNCTION(gnutls_session_get_data)
{

	zval *session_arg;
	gnutls_session_t *session;

	size_t session_data_size = 0;
	void *session_data;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_arg) == FAILURE) 
	{
	    return;
	}

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);

	gnutls_session_get_data (*session, NULL, &session_data_size);

	session_data = emalloc(session_data_size);
	gnutls_session_get_data (*session, session_data, &session_data_size);



	RETURN_STRINGL(session_data, session_data_size, 0);
}


ZEND_FUNCTION(gnutls_session_set_data)
{

	zval *session_arg;
	gnutls_session_t *session;
	char *session_data;
	int session_data_size;
	int retval;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &session_arg, &session_data, &session_data_size) == FAILURE) 
	{
	    return;
	}

	ZEND_FETCH_RESOURCE(session, gnutls_session_t *, &session_arg, -1, le_gnutls_session_name, le_gnutls_session);


	retval = gnutls_session_set_data (*session, session_data, session_data_size);


	RETURN_LONG(retval);
}
